import json
from pathlib import Path

from allotropy.constants import CHARDET_ENCODING
from allotropy.parser_factory import Vendor
from allotropy.to_allotrope import allotrope_from_file

output_files = [
    "96-Well Trevigen CometAssayCometChip Imaging and Analysis Sample File 23Nov15",
    "HeLa 96well Colony 11pt Doxorubicin UprBF_TAB",
    "Cell_Count_DAPI_GFP",
]
vendor = Vendor.AGILENT_GEN5_IMAGE


if __name__ == "__main__":
    # filename = "Beckman_Vi-Cell-XR_example02_instrumentOutput.xls"
    for filename in output_files:
        # read_mode = "fluorescence"
        # test_filepath = (
        #     f"tests/parsers/agilent_gen5/testdata/{read_mode}/{filename}.txt"
        # )
        # test_filepath = f"tests/parsers/luminex_xponent/testdata/{filename}.csv"
        test_filepath = f"{filename}.txt"

        allotrope_dict = allotrope_from_file(
            test_filepath, vendor, encoding=CHARDET_ENCODING
        )
        target_filename = Path(test_filepath).with_suffix(".json").name
        # print(allotrope_dict)
        with open(target_filename, "w") as fp:
            fp.write(json.dumps(allotrope_dict, indent=4, ensure_ascii=False))



from collections.abc import Sequence
from typing import Any

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    CalculatedDataAggregateDocument,
    CalculatedDataDocumentItem,
    ContainerType,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceSystemDocument,
    FluorescencePointDetectionDeviceControlAggregateDocument,
    FluorescencePointDetectionDeviceControlDocumentItem,
    FluorescencePointDetectionMeasurementDocumentItems,
    LuminescencePointDetectionDeviceControlAggregateDocument,
    LuminescencePointDetectionDeviceControlDocumentItem,
    LuminescencePointDetectionMeasurementDocumentItems,
    MeasurementAggregateDocument,
    Model,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    SampleDocument,
    UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument,
    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem,
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMillimeter,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueRelativeLightUnit,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_gen5.agilent_gen5_structure import PlateData
from allotropy.parsers.agilent_gen5.constants import (
    DEFAULT_SOFTWARE_NAME,
    DEVICE_TYPE,
    MULTIPLATE_FILE_ERROR,
    ReadMode,
)
from allotropy.parsers.agilent_gen5.section_reader import SectionLinesReader
from allotropy.parsers.lines_reader import read_to_lines
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser

MeasurementDocumentItems = (
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems
    | FluorescencePointDetectionMeasurementDocumentItems
    | LuminescencePointDetectionMeasurementDocumentItems
)

MeasurementDocumentAttributeClasses = (
    TQuantityValueDegreeCelsius
    | TQuantityValueMillimeter
    | TQuantityValueNanometer
    | TQuantityValueNumber
)


def get_instance_or_none(
    cls: type[MeasurementDocumentAttributeClasses], value: Any
) -> Any:
    return cls(value=value) if value is not None else None


class AgilentGen5Parser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Agilent Gen5"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def _create_model(self, plate_data: PlateData, file_name: str) -> Model:
        header_data = plate_data.header_data
        results = plate_data.results

        calculated_data_document = [
            CalculatedDataDocumentItem(
                calculated_data_identifier=calculated_datum.identifier,
                data_source_aggregate_document=DataSourceAggregateDocument(
                    data_source_document=[
                        DataSourceDocumentItem(
                            data_source_identifier=data_source.identifier,
                            data_source_feature=data_source.feature.value.lower(),
                        )
                        for data_source in calculated_datum.data_sources
                    ]
                ),
                calculated_data_name=calculated_datum.name,
                calculated_result=TQuantityValueUnitless(value=calculated_datum.result),
            )
            for calculated_datum in results.calculated_data
        ]

        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/plate-reader/BENCHLING/2023/09/plate-reader.manifest",
            plate_reader_aggregate_document=PlateReaderAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier="NA",
                    model_number=header_data.model_number,
                    equipment_serial_number=header_data.equipment_serial_number,
                ),
                data_system_document=DataSystemDocument(
                    file_name=file_name,
                    software_name=DEFAULT_SOFTWARE_NAME,
                    software_version=plate_data.header_data.software_version,
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                plate_reader_document=[
                    self._get_plate_reader_document_item(plate_data, well_position)
                    for well_position in results.wells
                ],
                calculated_data_aggregate_document=(
                    CalculatedDataAggregateDocument(
                        calculated_data_document=calculated_data_document
                    )
                    if calculated_data_document
                    else None
                ),
            ),
        )

    def _get_plate_reader_document_item(
        self, plate_data: PlateData, well_position: str
    ) -> PlateReaderDocumentItem:
        header_data = plate_data.header_data
        plate_well_count = len(plate_data.results.wells)
        read_mode = plate_data.read_data.read_mode

        measurement_document: Sequence[MeasurementDocumentItems]

        if read_mode == ReadMode.ABSORBANCE:
            measurement_document = self._get_absorbance_measurement_document(
                plate_data, well_position
            )
        elif read_mode == ReadMode.FLUORESCENCE:
            measurement_document = self._get_fluorescence_measurement_document(
                plate_data, well_position
            )
        elif read_mode == ReadMode.LUMINESCENCE:
            measurement_document = self._get_luminescence_measurement_document(
                plate_data, well_position
            )

        return PlateReaderDocumentItem(
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_time=self._get_date_time(header_data.datetime),
                analytical_method_identifier=header_data.protocol_file_path,
                experimental_data_identifier=header_data.experiment_file_path,
                plate_well_count=TQuantityValueNumber(value=plate_well_count),
                container_type=ContainerType.well_plate,
                measurement_document=list(measurement_document),
            )
        )

    def _get_sample_document(
        self, plate_data: PlateData, well_position: str
    ) -> SampleDocument:
        well_plate_identifier = plate_data.header_data.well_plate_identifier
        sample_identifier = plate_data.layout_data.sample_identifiers.get(
            well_position, f"{well_plate_identifier} {well_position}"
        )

        return SampleDocument(
            sample_identifier=sample_identifier,
            location_identifier=well_position,
            well_plate_identifier=well_plate_identifier,
        )

    def _get_wavelength_from_label(self, label: str) -> float:
        return float(label.split(":")[-1].split(" ")[0])

    def _get_absorbance_measurement_document(
        self, plate_data: PlateData, well_position: str
    ) -> list[UltravioletAbsorbancePointDetectionMeasurementDocumentItems]:
        read_data = plate_data.read_data

        measurements = plate_data.results.measurements[well_position]
        sample_document = self._get_sample_document(plate_data, well_position)

        return [
            UltravioletAbsorbancePointDetectionMeasurementDocumentItems(
                measurement_identifier=measurement.identifier,
                sample_document=sample_document,
                device_control_aggregate_document=UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument(
                    device_control_document=[
                        UltravioletAbsorbancePointDetectionDeviceControlDocumentItem(
                            device_type=DEVICE_TYPE,
                            detection_type=read_data.read_mode.value,
                            detector_wavelength_setting=TQuantityValueNanometer(
                                value=self._get_wavelength_from_label(measurement.label)
                            ),
                            number_of_averages=(
                                TQuantityValueNumber(value=read_data.number_of_averages)
                                if read_data.number_of_averages
                                else None
                            ),
                            detector_carriage_speed_setting=read_data.detector_carriage_speed,
                        )
                    ]
                ),
                absorbance=TQuantityValueMilliAbsorbanceUnit(value=measurement.value),
                compartment_temperature=get_instance_or_none(
                    TQuantityValueDegreeCelsius, plate_data.compartment_temperature
                ),
            )
            for measurement in measurements
        ]

    def _get_fluorescence_measurement_document(
        self, plate_data: PlateData, well_position: str
    ) -> list[FluorescencePointDetectionMeasurementDocumentItems]:
        read_data = plate_data.read_data
        measurements = plate_data.results.measurements[well_position]
        sample_document = self._get_sample_document(plate_data, well_position)

        measurement_document = []

        for measurement in measurements:
            filter_data = read_data.filter_sets[measurement.label]
            document = FluorescencePointDetectionMeasurementDocumentItems(
                measurement_identifier=measurement.identifier,
                sample_document=sample_document,
                device_control_aggregate_document=FluorescencePointDetectionDeviceControlAggregateDocument(
                    device_control_document=[
                        FluorescencePointDetectionDeviceControlDocumentItem(
                            device_type=DEVICE_TYPE,
                            detection_type=read_data.read_mode.value,
                            detector_wavelength_setting=get_instance_or_none(
                                TQuantityValueNanometer,
                                filter_data.detector_wavelength_setting,
                            ),
                            detector_bandwidth_setting=get_instance_or_none(
                                TQuantityValueNanometer,
                                filter_data.detector_bandwidth_setting,
                            ),
                            excitation_wavelength_setting=get_instance_or_none(
                                TQuantityValueNanometer,
                                filter_data.excitation_wavelength_setting,
                            ),
                            excitation_bandwidth_setting=get_instance_or_none(
                                TQuantityValueNanometer,
                                filter_data.excitation_bandwidth_setting,
                            ),
                            wavelength_filter_cutoff_setting=get_instance_or_none(
                                TQuantityValueNanometer,
                                filter_data.wavelength_filter_cutoff_setting,
                            ),
                            detector_distance_setting__plate_reader_=get_instance_or_none(
                                TQuantityValueMillimeter, read_data.detector_distance
                            ),
                            scan_position_setting__plate_reader_=filter_data.scan_position_setting,
                            detector_gain_setting=filter_data.gain,
                            number_of_averages=get_instance_or_none(
                                TQuantityValueNumber, read_data.number_of_averages
                            ),
                            detector_carriage_speed_setting=read_data.detector_carriage_speed,
                        )
                    ]
                ),
                fluorescence=TQuantityValueRelativeFluorescenceUnit(
                    value=measurement.value
                ),
                compartment_temperature=get_instance_or_none(
                    TQuantityValueDegreeCelsius, plate_data.compartment_temperature
                ),
            )
            measurement_document.append(document)

        return measurement_document

    def _get_luminescence_measurement_document(
        self, plate_data: PlateData, well_position: str
    ) -> list[LuminescencePointDetectionMeasurementDocumentItems]:
        _ = (plate_data, well_position)
        read_data = plate_data.read_data
        measurements = plate_data.results.measurements[well_position]
        sample_document = self._get_sample_document(plate_data, well_position)

        measurement_document = []
        for measurement in measurements:
            filter_data = read_data.filter_sets[measurement.label]
            measurement_document.append(
                LuminescencePointDetectionMeasurementDocumentItems(
                    measurement_identifier=measurement.identifier,
                    sample_document=sample_document,
                    device_control_aggregate_document=LuminescencePointDetectionDeviceControlAggregateDocument(
                        device_control_document=[
                            LuminescencePointDetectionDeviceControlDocumentItem(
                                device_type=DEVICE_TYPE,
                                detection_type=read_data.read_mode.value,
                                detector_wavelength_setting=get_instance_or_none(
                                    TQuantityValueNanometer,
                                    filter_data.detector_wavelength_setting,
                                ),
                                detector_bandwidth_setting=get_instance_or_none(
                                    TQuantityValueNanometer,
                                    filter_data.detector_bandwidth_setting,
                                ),
                                detector_distance_setting__plate_reader_=get_instance_or_none(
                                    TQuantityValueMillimeter,
                                    read_data.detector_distance,
                                ),
                                scan_position_setting__plate_reader_=filter_data.scan_position_setting,
                                detector_gain_setting=filter_data.gain,
                                detector_carriage_speed_setting=read_data.detector_carriage_speed,
                            )
                        ]
                    ),
                    luminescence=TQuantityValueRelativeLightUnit(
                        value=measurement.value
                    ),
                )
            )
        return measurement_document

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Any:
        lines = read_to_lines(named_file_contents)
        section_lines_reader = SectionLinesReader(lines)
        plates = list(section_lines_reader.iter_sections("^Software Version"))

        if not plates:
            msg = "No plate data found in file."
            raise AllotropeConversionError(msg)

        if len(plates) > 1:
            raise AllotropeConversionError(MULTIPLATE_FILE_ERROR)

        plate_data = PlateData.create(plates[0], named_file_contents.original_file_name)

        return self._create_model(plate_data, named_file_contents.original_file_name)



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from pathlib import PureWindowsPath
import re
from typing import Optional

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.values import assert_not_none, try_float_or_none


@dataclass(frozen=True)
class DeviceInfo:
    device_identifier: str
    model_number: str
    equipment_serial_number: str
    data_system_instance_id: str
    basename: str
    file_name: str
    unc_path: str
    software_name: str
    software_version: str
    measurement_time: str
    analyst: str

    @staticmethod
    def create(reader: LinesReader) -> DeviceInfo:
        data = {}
        for i in range(3):  # for first three lines
            line = assert_not_none(
                reader.pop(),
                msg=f"Unable to read line number {i+1}",
            )
            for element in line.split(";"):
                parts = element.split(":", maxsplit=1)
                if len(parts) > 1:
                    data[parts[0].strip()] = parts[1].strip()

        raw_path = assert_not_none(
            reader.pop(),
            msg="Unable to read file path",
        )

        path = PureWindowsPath(raw_path)

        reader.drop_until_empty()
        reader.drop_empty()

        analyzer_serial_number = assert_not_none(
            data.get("Analyzer Serial number"),
            msg="Unable to find analyzer serial number.",
        )

        analyzer_serial_match = assert_not_none(
            re.match(r"^(\w+)-(\w+)", analyzer_serial_number),
            msg="Unable to parse analyzer serial number.",
        )

        computer_name = assert_not_none(
            data.get("Computer name"),
            msg="Unable to find computer name",
        )

        software_info = assert_not_none(
            data.get("Software version"),
            msg="Unable to find software version",
        )

        software_info_match = assert_not_none(
            re.match(r"^ImmunoSpot ([\d\.]+)$", software_info),
            msg="Unable to parse software version",
        )

        counted = assert_not_none(
            data.get("Counted"),
            msg="Unable to find counted timestamp.",
        )

        authenticated_user = assert_not_none(
            data.get("Authenticated user"),
            msg="Unable to find authenticated user.",
        )

        return DeviceInfo(
            device_identifier="N/A",
            model_number=analyzer_serial_match.group(1),
            equipment_serial_number=analyzer_serial_number,
            data_system_instance_id=computer_name,
            basename=path.name,
            file_name=path.stem,
            unc_path=raw_path,
            software_name="ImmunoSpot",
            software_version=software_info_match.group(1),
            measurement_time=counted,
            analyst=authenticated_user,
        )


@dataclass(frozen=True)
class Well:
    col: str
    row: str
    value: Optional[float]

    @property
    def pos(self) -> str:
        return self.row + self.col


@dataclass(frozen=True)
class Plate:
    name: str
    wells: dict[str, Well]

    @staticmethod
    def create(reader: LinesReader) -> Plate:
        raw_name = assert_not_none(
            reader.pop(),
            msg="Unable to read name of assay data.",
        )

        name_match = assert_not_none(
            re.search("\t\t\t(.+)", raw_name),
            msg="Unable to parse name of assay data.",
        )
        name = re.sub(r"\s+", " ", name_match.group(1))

        if "ImmunoSpot Plate Code" in name:
            name = "Spot Count"

        raw_columns = assert_not_none(
            reader.pop(),
            msg=f"Unable to read data from {name}",
        )
        columns = re.sub(r"\s+", " ", raw_columns.strip())

        reader.drop_empty()

        wells = {}
        for raw_line in reader.pop_until_empty():
            line = re.sub(r"\s+", " ", raw_line)
            if match := re.match(r"^([A-Z])\s+(.+)", line):
                raw_values = match.group(2).strip()
                for column, value in zip(columns.split(), raw_values.split()):
                    well = Well(
                        col=column,
                        row=match.group(1),
                        value=try_float_or_none(value),
                    )
                    wells[well.pos] = well

        return Plate(name, wells)

    def get_well_count(self) -> int:
        return len(self.wells)

    def iter_wells(self) -> Iterator[Well]:
        yield from self.wells.values()

    def get_well(self, pos: str) -> Well:
        return assert_not_none(
            self.wells.get(pos),
            msg=f"Unable to find well '{pos}' in plate '{self.name}'",
        )


@dataclass(frozen=True)
class AssayData:
    plates: list[Plate]
    identifier: Optional[str]
    well_count: int

    @staticmethod
    def create(reader: LinesReader) -> AssayData:
        reader.drop_until_inclusive("Unprocessed Data$")

        reader.drop_empty()
        plate_code_line = reader.get() or ""
        assert_not_none(
            re.search("Plate Code =", plate_code_line),
            msg="Unable to find ImmunoSpot Plate Code line",
        )

        identifier = (
            match.group(1)
            if (match := re.search(r"Plate Code = ([\w ]+)", plate_code_line))
            else None
        )

        plates = []
        while reader.current_line_exists():
            reader.drop_empty()
            plates.append(Plate.create(reader))
            reader.drop_empty()

        if not plates:
            error = "Unable to find plate information."
            raise AllotropeConversionError(error)

        return AssayData(plates, identifier, well_count=plates[0].get_well_count())

    def iter_wells(self) -> Iterator[Well]:
        yield from self.plates[0].iter_wells()

    def iter_plates_well(self, pos: str) -> Iterator[Well]:
        for plate in self.plates:
            yield plate.get_well(pos)


@dataclass(frozen=True)
class Data:
    device_info: DeviceInfo
    assay_data: AssayData

    @staticmethod
    def create(reader: LinesReader) -> Data:
        device_info = DeviceInfo.create(reader)

        reader.drop_empty()
        reader.drop_until_empty()  # ignore assay info
        reader.drop_empty()

        assay_data = AssayData.create(reader)
        return Data(device_info, assay_data)

    def get_plate_identifier(self) -> str:
        return (
            self.device_info.file_name
            if self.assay_data.identifier is None
            else self.assay_data.identifier
        )



from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    ContainerType,
    DataSystemDocument,
    DeviceSystemDocument,
    ImageFeatureAggregateDocument,
    ImageFeatureDocumentItem,
    MeasurementAggregateDocument,
    Model,
    OpticalImagingDeviceControlAggregateDocument,
    OpticalImagingDeviceControlDocumentItem,
    OpticalImagingMeasurementDocumentItems,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueNumber,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    InvalidJsonFloat,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.ctl_immunospot.ctl_immunospot_structure import Data, Well
from allotropy.parsers.lines_reader import LinesReader, read_to_lines
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser


class CtlImmunospotParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "CTL ImmunoSpot"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        lines = read_to_lines(named_file_contents)
        reader = LinesReader(lines)
        data = Data.create(reader)
        return self._get_model(data)

    def _get_model(self, data: Data) -> Model:
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/plate-reader/BENCHLING/2023/09/plate-reader.manifest",
            plate_reader_aggregate_document=self._get_plate_reader_aggregate_document(
                data
            ),
        )

    def _get_plate_reader_aggregate_document(
        self, data: Data
    ) -> PlateReaderAggregateDocument:
        return PlateReaderAggregateDocument(
            device_system_document=self._get_device_system_document(data),
            data_system_document=self._get_data_system_document(data),
            plate_reader_document=[
                self._get_plate_reader_document_item(data, well)
                for well in data.assay_data.iter_wells()
            ],
        )

    def _get_device_system_document(self, data: Data) -> DeviceSystemDocument:
        return DeviceSystemDocument(
            device_identifier=data.device_info.device_identifier,
            model_number=data.device_info.model_number,
            product_manufacturer="CTL",
            equipment_serial_number=data.device_info.equipment_serial_number,
        )

    def _get_data_system_document(self, data: Data) -> DataSystemDocument:
        return DataSystemDocument(
            data_system_instance_identifier=data.device_info.data_system_instance_id,
            file_name=data.device_info.basename,
            UNC_path=data.device_info.unc_path,
            software_name=data.device_info.software_name,
            software_version=data.device_info.software_version,
            ASM_converter_name=ASM_CONVERTER_NAME,
            ASM_converter_version=ASM_CONVERTER_VERSION,
        )

    def _get_plate_reader_document_item(
        self, data: Data, well: Well
    ) -> PlateReaderDocumentItem:
        return PlateReaderDocumentItem(
            analyst=data.device_info.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_time=self._get_date_time(data.device_info.measurement_time),
                container_type=ContainerType.well_plate,
                plate_well_count=TQuantityValueNumber(value=data.assay_data.well_count),
                measurement_document=[self._get_measurement_document(data, well)],
            ),
        )

    def _get_measurement_document(
        self, data: Data, well: Well
    ) -> OpticalImagingMeasurementDocumentItems:
        return OpticalImagingMeasurementDocumentItems(
            measurement_identifier=random_uuid_str(),
            device_control_aggregate_document=OpticalImagingDeviceControlAggregateDocument(
                device_control_document=[
                    OpticalImagingDeviceControlDocumentItem(
                        device_type="imager",
                        detection_type="optical-imaging",
                    )
                ]
            ),
            sample_document=self._get_sample_document(data, well),
            processed_data_aggregate_document=self._get_processed_data_aggregate_document(
                data, well
            ),
        )

    def _get_sample_document(self, data: Data, well: Well) -> SampleDocument:
        well_plate_identifier = data.get_plate_identifier()
        return SampleDocument(
            sample_identifier=f"{well_plate_identifier}_{well.pos}",
            location_identifier=well.pos,
            well_plate_identifier=well_plate_identifier,
        )

    def _get_processed_data_aggregate_document(
        self, data: Data, well: Well
    ) -> ProcessedDataAggregateDocument:
        return ProcessedDataAggregateDocument(
            processed_data_document=[
                ProcessedDataDocumentItem(
                    processed_data_identifier=random_uuid_str(),
                    image_feature_aggregate_document=ImageFeatureAggregateDocument(
                        image_feature_document=[
                            ImageFeatureDocumentItem(
                                image_feature_identifier=random_uuid_str(),
                                image_feature_name=plate.name,
                                image_feature_result=TQuantityValueUnitless(
                                    value=plate.get_well(well.pos).value
                                    or InvalidJsonFloat.NaN
                                ),
                            )
                            for plate in data.assay_data.plates
                        ],
                    ),
                )
            ]
        )



import re
from typing import Any, Optional, Union

import pandas as pd

from allotropy.allotrope.models.adm.light_obscuration.benchling._2023._12.light_obscuration import (
    CalculatedDataDocumentItem,
    DataProcessingDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceDocumentItem,
    DeviceSystemDocument,
    DistributionAggregateDocument,
    DistributionDocumentItem,
    DistributionItem,
    LightObscurationAggregateDocument,
    LightObscurationDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    TCalculatedDataAggregateDocument,
    TDataSourceAggregateDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCountsPerMilliliter,
    TQuantityValueMicrometer,
    TQuantityValueMilliliter,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import TStringValueItem
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_pharmspec.constants import PHARMSPEC_SOFTWARE_NAME
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser

# This map is used to coerce the column names coming in the raw data
# into names of the allotrope properties.
column_map = {
    "Cumulative Counts/mL": "cumulative particle density",
    "Cumulative Count": "cumulative count",
    "Particle Size(µm)": "particle size",
    "Differential Counts/mL": "differential particle density",
    "Differential Count": "differential count",
}

property_lookup = {
    "particle_size": TQuantityValueMicrometer,
    "cumulative_count": TQuantityValueUnitless,
    "cumulative_particle_density": TQuantityValueCountsPerMilliliter,
    "differential_particle_density": TQuantityValueCountsPerMilliliter,
    "differential_count": TQuantityValueUnitless,
}

VALID_CALCS = ["Average"]


def get_property_from_sample(property_name: str, value: Any) -> Any:
    return property_lookup[property_name](value=value)


class PharmSpecParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Beckman PharmSpec"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        df = pd.read_excel(named_file_contents.contents, header=None, engine="openpyxl")
        return self._setup_model(df, named_file_contents.original_file_name)

    def _get_data_using_key_bounds(
        self, df: pd.DataFrame, start_key: str, end_key: str
    ) -> pd.DataFrame:
        """Find the data in the raw dataframe. We identify the boundary of the data
        by finding the index first row which contains the word 'Particle' and ending right before
        the index of the first row containing 'Approver'.

        :param df: the raw dataframe
        :param start_key: the key to start the slice
        :parm end_key: the key to end the slice
        :return: the dataframe slice between the stard and end bounds
        """
        start = df[df[1].str.contains(start_key, na=False)].index.values[0]
        end = df[df[0].str.contains(end_key, na=False)].index.values[0] - 1
        return df.loc[start:end, :]

    def _extract_data(self, df: pd.DataFrame) -> pd.DataFrame:
        """Extract the Average data frame from the raw data. Initial use cases have focused on
        only extracting the Average data, not the individual runs. The ASM does support multiple
        Distribution objects, but they don't have names, so it's not possible to pick these out
        after the fact. As such, this extraction only includes the Average data.

        :param df: the raw dataframe
        :return: the average data frame
        """
        data = self._get_data_using_key_bounds(
            df, start_key="Particle", end_key="Approver_"
        )
        data = data.dropna(how="all").dropna(how="all", axis=1)
        data[0] = data[0].ffill()
        data = data.dropna(subset=1).reset_index(drop=True)
        data.columns = pd.Index([x.strip() for x in data.loc[0]])
        data = data.loc[1:, :]
        return data.rename(columns={x: column_map[x] for x in column_map})

    def _create_distribution_document_items(
        self, df: pd.DataFrame
    ) -> list[DistributionDocumentItem]:
        """Create the distribution document. First, we create the actual distrituion, which itself
        contains a list of DistributionDocumentItem objects. The DistributionDocumentItem objects represent the values
        from the rows of the incoming dataframe.

        If we were able to support more than one data frame instead of just the average data, we could
        return a DistributionDocument with more than one item. For the use cases we've seen, there is
        only a single Distribution being returned at this time, containing the average data.

        :param df: The average dataframe
        :return: The DistributionDocument
        """
        cols = [v for k, v in column_map.items()]
        items = []
        for elem in df.to_dict("records"):
            item = {}
            for c in cols:
                prop = c.replace(
                    " ", "_"
                )  # to be able to set the props on the DistributionItem
                if c in elem:
                    item[prop] = get_property_from_sample(prop, float(elem[c]))
            item["distribution_identifier"] = random_uuid_str()
            items.append(DistributionItem(**item))
        # TODO get test example for data_processing_omission_setting
        dd = DistributionDocumentItem(distribution=items)
        return [dd]

    def _create_calculated_document_items(
        self, df: pd.DataFrame, feature: str, run_names: list[str]
    ) -> list[CalculatedDataDocumentItem]:
        cols = column_map.values()
        items = []
        for row in df.index:
            particle_size = df.at[
                row, "particle size"
            ]  # to track the calcuated data tuple
            for col in [x for x in cols if x in df.columns]:
                prop = col.replace(
                    " ", "_"
                )  # to be able to set the props on the DistributionItem
                items.append(
                    CalculatedDataDocumentItem(
                        calculated_data_identifier=random_uuid_str(),
                        calculated_data_name=f"{feature}_{prop}".lower(),
                        calculated_result=get_property_from_sample(
                            prop, df.at[row, col]
                        ),
                        data_source_aggregate_document=TDataSourceAggregateDocument(
                            data_source_document=[
                                DataSourceDocumentItem(
                                    data_source_identifier=f"{run_name}|{particle_size}",  # will be replaced by distribution id
                                    data_source_feature=col,
                                )
                                for run_name in run_names
                                if run_name not in VALID_CALCS
                            ]
                        ),
                    )
                )
        return items

    def _get_software_version_report_string(self, report_string: str) -> str:
        match = re.search(r"v(\d+(?:\.\d+)?(?:\.\d+)?)", report_string)
        if match:
            return match.group(1)
        return "Unknown"

    def _create_calculated_data_aggregate_document(
        self, df: pd.DataFrame, name: str, run_names: list[str]
    ) -> TCalculatedDataAggregateDocument:
        return TCalculatedDataAggregateDocument(
            calculated_data_document=self._create_calculated_document_items(
                df, name, run_names
            )
        )

    def _create_measurement_document_item(
        self, df: pd.DataFrame, gdf: pd.DataFrame, name: str
    ) -> MeasurementDocumentItem:
        return MeasurementDocumentItem(
            measurement_identifier=name,
            measurement_time=pd.to_datetime(
                str(df.at[8, 5]).replace(".", "-")
            ).isoformat(timespec="microseconds")
            + "Z",
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(
                        flush_volume_setting=TQuantityValueMilliliter(value=0),
                        detector_view_volume=TQuantityValueMilliliter(
                            value=df.at[9, 5]
                        ),
                        repetition_setting=int(df.at[11, 5]),
                        sample_volume_setting=TQuantityValueMilliliter(
                            value=df.at[11, 2]
                        ),
                    )
                ]
            ),
            sample_document=SampleDocument(
                sample_identifier=str(df.at[2, 2]),
            ),
            processed_data_aggregate_document=ProcessedDataAggregateDocument(
                processed_data_document=[
                    ProcessedDataDocumentItem(
                        data_processing_document=DataProcessingDocument(
                            dilution_factor_setting=TQuantityValueUnitless(
                                value=df.at[13, 2]
                            ),
                        ),
                        distribution_aggregate_document=DistributionAggregateDocument(
                            distribution_document=self._create_distribution_document_items(
                                gdf
                            )
                        ),
                    )
                ]
            ),
        )

    def _create_model(
        self,
        df: pd.DataFrame,
        calc_agg_doc: Optional[TCalculatedDataAggregateDocument],
        measurement_doc_items: list[MeasurementDocumentItem],
        file_name: str,
    ) -> Model:
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/light-obscuration/BENCHLING/2023/12/light-obscuration.manifest",
            light_obscuration_aggregate_document=LightObscurationAggregateDocument(
                light_obscuration_document=[
                    LightObscurationDocumentItem(
                        measurement_aggregate_document=MeasurementAggregateDocument(
                            analyst=str(df.at[6, 5]),
                            submitter=None,
                            measurement_document=measurement_doc_items,
                        )
                    )
                ],
                data_system_document=DataSystemDocument(
                    file_name=file_name,
                    software_name=PHARMSPEC_SOFTWARE_NAME,
                    software_version=self._get_software_version_report_string(
                        df.at[0, 2]
                    ),
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                device_system_document=DeviceSystemDocument(
                    equipment_serial_number=str(df.at[4, 5]),
                    device_document=[
                        DeviceDocumentItem(
                            detector_identifier="",
                            detector_model_number=str(df.at[2, 5]),
                        )
                    ],
                ),
                calculated_data_aggregate_document=calc_agg_doc,
            ),
        )

    def _get_distribution_id_for_run_and_particle_size(
        self,
        run_name: str,
        particle_size: float,
        measurement_doc_items: list[MeasurementDocumentItem],
    ) -> Optional[Union[str, TStringValueItem]]:
        item = next(
            x for x in measurement_doc_items if x.measurement_identifier == run_name
        )
        if (
            item.processed_data_aggregate_document is not None
            and item.processed_data_aggregate_document.processed_data_document
            is not None
        ):
            for pdd in item.processed_data_aggregate_document.processed_data_document:
                if (
                    pdd.distribution_aggregate_document is not None
                    and pdd.distribution_aggregate_document.distribution_document
                    is not None
                ):
                    for dd in pdd.distribution_aggregate_document.distribution_document:
                        for d in dd.distribution:
                            if d.particle_size.value == particle_size:
                                return d.distribution_identifier
        return None

    def _add_data_source_to_calculated_data(
        self,
        calc_agg_doc: TCalculatedDataAggregateDocument,
        measurement_doc_items: list[MeasurementDocumentItem],
    ) -> None:
        if calc_agg_doc.calculated_data_document is not None:
            for cdd in calc_agg_doc.calculated_data_document:
                if (
                    cdd.data_source_aggregate_document is not None
                    and cdd.data_source_aggregate_document.data_source_document
                    is not None
                ):
                    for dsd in cdd.data_source_aggregate_document.data_source_document:
                        run_name, particle_size = str(dsd.data_source_identifier).split(
                            "|"
                        )
                        distribution_id = (
                            self._get_distribution_id_for_run_and_particle_size(
                                run_name, float(particle_size), measurement_doc_items
                            )
                        )
                        if distribution_id is not None:
                            dsd.data_source_identifier = distribution_id

    def _setup_model(self, df: pd.DataFrame, file_name: str) -> Model:
        """Build the Model

        :param df: the raw dataframe
        :return: the model
        """
        data = self._extract_data(df)
        measurement_doc_items = []
        calc_agg_doc = None
        run_names = list(data["Run No."].unique())
        for g, gdf in data.groupby("Run No."):
            name = str(g)
            if g in VALID_CALCS:
                calc_agg_doc = self._create_calculated_data_aggregate_document(
                    gdf, name, run_names
                )
            else:
                measurement_doc_items.append(
                    self._create_measurement_document_item(df, gdf, name)
                )
        if calc_agg_doc is not None:
            self._add_data_source_to_calculated_data(
                calc_agg_doc, measurement_doc_items
            )
        return self._create_model(df, calc_agg_doc, measurement_doc_items, file_name)



from collections.abc import Sequence
from typing import Any, Union

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    ContainerType,
    DataSystemDocument,
    DeviceSystemDocument,
    ImageFeatureAggregateDocument,
    ImageFeatureDocumentItem,
    MeasurementAggregateDocument,
    Model,
    OpticalImagingDeviceControlAggregateDocument,
    OpticalImagingDeviceControlDocumentItem,
    OpticalImagingMeasurementDocumentItems,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMilliSecond,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_gen5.section_reader import SectionLinesReader
from allotropy.parsers.agilent_gen5_image.agilent_gen5_image_structure import (
    ImageFeature,
    PlateData,
    ReadData,
)
from allotropy.parsers.agilent_gen5_image.constants import (
    DEFAULT_SOFTWARE_NAME,
    DETECTION_TYPE,
    DEVICE_TYPE,
    MULTIPLATE_FILE_ERROR,
    NO_PLATE_DATA_ERROR,
)
from allotropy.parsers.lines_reader import read_to_lines
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser

MeasurementDocumentAttributeClasses = Union[
    TQuantityValueMilliSecond,
    TQuantityValueNanometer,
    TQuantityValueUnitless,
]


def get_instance_or_none(
    cls: type[MeasurementDocumentAttributeClasses], value: Any
) -> Any:
    return cls(value=value) if value is not None else None


class AgilentGen5ImageParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Agilent Gen5 Image"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def _create_model(self, plate_data: PlateData, file_name: str) -> Model:
        header_data = plate_data.header_data

        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/plate-reader/BENCHLING/2023/09/plate-reader.manifest",
            plate_reader_aggregate_document=PlateReaderAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier="N/A",
                    model_number=header_data.model_number,
                    equipment_serial_number=header_data.equipment_serial_number,
                ),
                data_system_document=DataSystemDocument(
                    file_name=file_name,
                    software_name=DEFAULT_SOFTWARE_NAME,
                    software_version=plate_data.header_data.software_version,
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                plate_reader_document=[
                    self._get_plate_reader_document_item(plate_data, well_position)
                    for well_position in plate_data.results.image_features.keys()
                ],
            ),
        )

    def _get_plate_reader_document_item(
        self, plate_data: PlateData, well_position: str
    ) -> PlateReaderDocumentItem:
        header_data = plate_data.header_data
        plate_well_count = len(plate_data.results.image_features)

        image_features = plate_data.results.image_features[well_position]
        processed_data_aggregate_document = ProcessedDataAggregateDocument(
            processed_data_document=self._get_processed_data_document(image_features)
        )

        sample_document = self._get_sample_document(plate_data, well_position)
        measurement_document = list(
            self._get_measurement_document(plate_data.read_data, sample_document)
        )

        # Image features (included in the processed data document) are included at the measurement
        # level only when there is only one device control document (and thus one measurement document)
        # otherwise they are included at the measurement aggregate document level.
        if len(measurement_document) == 1:
            first_doc = measurement_document[0]
            first_doc.processed_data_aggregate_document = (
                processed_data_aggregate_document
            )

        return PlateReaderDocumentItem(
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_time=self._get_date_time(header_data.datetime),
                analytical_method_identifier=header_data.protocol_file_path,
                experimental_data_identifier=header_data.experiment_file_path,
                plate_well_count=TQuantityValueNumber(value=plate_well_count),
                container_type=ContainerType.well_plate,
                measurement_document=list(measurement_document),
                processed_data_aggregate_document=(
                    processed_data_aggregate_document
                    if len(measurement_document) > 1
                    else None
                ),
            )
        )

    def _get_processed_data_document(
        self,
        image_features: list[ImageFeature],
    ) -> list[ProcessedDataDocumentItem]:
        return [
            ProcessedDataDocumentItem(
                processed_data_identifier=random_uuid_str(),
                image_feature_aggregate_document=ImageFeatureAggregateDocument(
                    image_feature_document=[
                        ImageFeatureDocumentItem(
                            image_feature_identifier=image_feature.identifier,
                            image_feature_name=image_feature.name,
                            image_feature_result=TQuantityValueUnitless(
                                value=image_feature.result
                            ),
                        )
                        for image_feature in image_features
                    ]
                ),
            )
        ]

    def _get_sample_document(
        self, plate_data: PlateData, well_position: str
    ) -> SampleDocument:
        well_plate_identifier = plate_data.header_data.well_plate_identifier
        sample_identifier = plate_data.layout_data.sample_identifiers.get(
            well_position, f"{well_plate_identifier} {well_position}"
        )

        return SampleDocument(
            sample_identifier=sample_identifier,
            location_identifier=well_position,
            well_plate_identifier=well_plate_identifier,
        )

    def _get_measurement_document(
        self, read_data: ReadData, sample_document: SampleDocument
    ) -> Sequence[OpticalImagingMeasurementDocumentItems]:
        measurement_document = []
        for read_section in read_data.read_sections:
            measurement_document.extend(
                [
                    OpticalImagingMeasurementDocumentItems(
                        measurement_identifier=random_uuid_str(),
                        sample_document=sample_document,
                        device_control_aggregate_document=OpticalImagingDeviceControlAggregateDocument(
                            device_control_document=[
                                OpticalImagingDeviceControlDocumentItem(
                                    device_type=DEVICE_TYPE,
                                    detection_type=DETECTION_TYPE,
                                    # This setting won't get reported at the moment since Gen5 only reports it
                                    # in micrometers and we don't do conversions on the adapters at the moment
                                    # detector_distance_setting__plate_reader_=get_instance_or_none(
                                    #     TQuantityValueMillimeter,
                                    #     instrument_settings.detector_distance,
                                    # ),
                                    detector_gain_setting=instrument_settings.detector_gain,
                                    magnification_setting=get_instance_or_none(
                                        TQuantityValueUnitless,
                                        read_section.magnification_setting,
                                    ),
                                    illumination_setting=get_instance_or_none(
                                        TQuantityValueUnitless,
                                        instrument_settings.illumination,
                                    ),
                                    transmitted_light_setting=instrument_settings.transmitted_light,
                                    auto_focus_setting=instrument_settings.auto_focus,
                                    image_count_setting=get_instance_or_none(
                                        TQuantityValueUnitless,
                                        read_section.image_count_setting,
                                    ),
                                    fluorescent_tag_setting=instrument_settings.fluorescent_tag,
                                    exposure_duration_setting=get_instance_or_none(
                                        TQuantityValueMilliSecond,
                                        instrument_settings.exposure_duration,
                                    ),
                                    excitation_wavelength_setting=get_instance_or_none(
                                        TQuantityValueNanometer,
                                        instrument_settings.excitation_wavelength,
                                    ),
                                    detector_wavelength_setting=get_instance_or_none(
                                        TQuantityValueNanometer,
                                        instrument_settings.detector_wavelength,
                                    ),
                                )
                            ]
                        ),
                    )
                    for instrument_settings in read_section.instrument_settings_list
                ]
            )

        return measurement_document

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Any:
        lines = read_to_lines(named_file_contents)
        section_lines_reader = SectionLinesReader(lines)
        plates = list(section_lines_reader.iter_sections("^Software Version"))

        if not plates:
            raise AllotropeConversionError(NO_PLATE_DATA_ERROR)

        if len(plates) > 1:
            raise AllotropeConversionError(MULTIPLATE_FILE_ERROR)

        plate_data = PlateData.create(plates[0], named_file_contents.original_file_name)

        return self._create_model(plate_data, named_file_contents.original_file_name)



from collections.abc import Mapping
import importlib
from pathlib import Path
import re
from typing import Any

ALLOTROPE_DIR: Path = Path(__file__).parent.parent
SCHEMA_DIR_PATH: Path = Path(ALLOTROPE_DIR, "schemas")
SHARED_SCHEMAS_PATH: Path = Path(SCHEMA_DIR_PATH, "shared")
SHARED_SCHEMAS_DEFINITIONS_PATH: Path = Path(SHARED_SCHEMAS_PATH, "definitions")
UNITS_SCHEMAS_PATH: Path = Path(SHARED_SCHEMAS_DEFINITIONS_PATH, "units.json")
CUSTOM_SCHEMAS_PATH: Path = Path(SHARED_SCHEMAS_DEFINITIONS_PATH, "custom.json")
MODEL_DIR_PATH: Path = Path(ALLOTROPE_DIR, "models")
SHARED_MODELS_PATH: Path = Path(MODEL_DIR_PATH, "shared")
SHARED_MODELS_DEFINITIONS_PATH: Path = Path(SHARED_MODELS_PATH, "definitions")
UNITS_MODELS_PATH: Path = Path(SHARED_MODELS_DEFINITIONS_PATH, "units.py")
CUSTOM_MODELS_PATH: Path = Path(SHARED_MODELS_DEFINITIONS_PATH, "custom.py")
GENERATED_SHARED_PATHS: list[Path] = [
    UNITS_SCHEMAS_PATH,
    UNITS_MODELS_PATH,
    CUSTOM_SCHEMAS_PATH,
    CUSTOM_MODELS_PATH,
]


def get_rel_schema_path(schema_path: Path) -> Path:
    try:
        return schema_path.relative_to(SCHEMA_DIR_PATH)
    except ValueError as err:
        if not Path(SCHEMA_DIR_PATH, schema_path).exists():
            msg = f"Invalid schema path: {schema_path}"
            raise ValueError(msg) from err
        return schema_path


def get_full_schema_path(schema_path: Path) -> Path:
    if str(schema_path).startswith(str(SCHEMA_DIR_PATH)):
        return schema_path
    elif not Path(SCHEMA_DIR_PATH, schema_path).exists():
        msg = f"Invalid schema path: {schema_path}"
        raise AssertionError(msg)
    return Path(SCHEMA_DIR_PATH, schema_path)


def get_manifest_from_schema_path(schema_path: Path) -> str:
    rel_schema_path = get_rel_schema_path(schema_path)
    if not str(rel_schema_path).startswith("adm/") or not str(rel_schema_path).endswith(
        ".schema.json"
    ):
        msg = f"Invalid schema path: {rel_schema_path}"
        raise ValueError(msg)
    return f"http://purl.allotrope.org/manifests/{str(rel_schema_path)[4:-12]}.manifest"


def get_schema_path_from_manifest(manifest: str) -> Path:
    match = re.match(r"http://purl.allotrope.org/manifests/(.*)\.manifest", manifest)
    if not match:
        msg = f"No matching schema in repo for manifest: {manifest}"
        raise ValueError(msg)
    return Path(f"adm/{match.groups()[0]}.schema.json")


def get_schema_path_from_reference(reference: str) -> Path:
    ref_match = re.match(r"http://purl.allotrope.org/json-schemas/(.*)", reference)
    if not ref_match:
        msg = f"Could not parse reference: {reference}"
        raise ValueError(msg)
    return Path(f"{ref_match.groups()[0]}.json")


def get_model_file_from_schema_path(schema_path: Path) -> Path:
    rel_schema_path = get_rel_schema_path(schema_path)
    schema_file = rel_schema_path.name
    model_file = schema_file.replace(".schema.json", ".py").replace("-", "_").lower()
    schema_dir = str(rel_schema_path.parent)
    model_path = re.sub("/([0-9]+)", r"/_\1", schema_dir.replace("-", "_").lower())
    return Path(model_path, model_file)


def get_model_class_from_schema(asm: Mapping[str, Any]) -> Any:
    schema_path = get_schema_path_from_manifest(asm["$asm.manifest"])
    model_file = get_model_file_from_schema_path(Path(schema_path))
    import_path = f"allotropy.allotrope.models.{str(model_file).replace('/', '.')[:-3]}"
    # NOTE: it is safe to assume that every schema module has Model, as we generate this code.
    return importlib.import_module(import_path).Model
